#include "twi-all-included.h"

//We use 10 bits I2C address but Atmega is only compatible with 7 bits address. That's not a problem. 
//We just have to trick a little bit because 10 bits addressing is completely fine with the standard I2C physical protocol

//Let's say address = [9876543210] (10 bits)

//Here is how to write communication to a 10bits addressed slave : (A = ACK)
//[START] [0b1111 098W]A [76543210]A [WRITE_1]A [WRITE_2]A ... [WRITE_N]A [STOP]

//We begin with the template '1111 0' in order to talk to 10-bits addressed slaves, followed by 2 address bits, and W.
//Then the first byte we transmit are the 8 following bits of the address, in order to finish the 10 bits addressing.
//Nothing else changes. In order to know if the slave is missing, it's the second ACK that should be checked.


//Read communication (NA = NACK)
//[START] [0b1111 098W]A [76543210]A [RESTART] [0b1111 098R]A [ANSWER_1]A [ANSWER_2]A ... [ANSWER_N]NA [STOP]
//We begin as a standard 10 bits addressed write in order to ask our question to the slave, then we do a restart in order to get the answer

//After the restart, only one byte can be placed (with Read Bit at the end) before reading, so we cant put the full 10 bits address. We put only but the begining of the address.
//The fully adressed component just before the restart will be the only one to answer.



uint16_t crc16(const uint8_t* data_p, uint8_t length)
{
    uint8_t x;
    uint16_t crc = 0xFFFF;

    while (length--)
    {
        x = crc >> 8 ^ *data_p++;
        x ^= x >> 4;
        crc = (crc << 8) ^ ((uint16_t)(x << 12)) ^ ((uint16_t)(x <<5)) ^ ((uint16_t)x);
    }
    return crc;
}

void setup()
{
  Serial.begin(9600); //[Start+8bits+STOP] => 960 bytes/s
  Serial.println("----- I2C Reset of the TI046B1 CHIPS ------");
  Serial.println("-------------------------------------------");
  
  twi_init(); //Standard 100 kHz Atmel's internal I2C device initialisation with Pull-Up on SDA/SCL
}

bool read_TI046B1_register(uint16_t TenBits_slave_address, uint8_t firstRegisterByte, uint8_t secondRegisterByte, uint8_t thirdRegisterByte, uint8_t * destinationBuffer, uint16_t readSize)
{
  uint16_t slave_address_LSB = TenBits_slave_address & 0x0FF; //0076543210 //8 LSB
  uint16_t slave_address_MSB = TenBits_slave_address & 0x300; //9800000000 //2 MSB
  
  //Put the MSB bits to the Left :
  slave_address_MSB = slave_address_MSB >> 8; //For example, 0x300 becomes 0x003
  
  //7 bits address equivalent for the begining :
  uint8_t SevenBits_compat_address = 0x78 | slave_address_MSB; //TWI library will put send those 7 bits followed by read/write bit.
  
  
  //Preparation of the write buffer : 8LSB of the slave's address, then 3 bytes writes to call a register read.
  uint8_t txBuffer[4];
  txBuffer[0] = slave_address_LSB;
  txBuffer[1] = firstRegisterByte;
  txBuffer[2] = secondRegisterByte;
  txBuffer[3] = thirdRegisterByte;
  
  //Execution of the 4 bytes write
  uint16_t nRet = twi_writeTo(SevenBits_compat_address, txBuffer, 4, 1, false); //4 bytes, wait, and no stop, since we will do a restart in order to read just after
  if (nRet == 1)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("Error occured wile trying to write to the slave : TWI class's buffer is too small");
    return false;
  }

  else if (nRet == 2)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("Error occured wile trying to write to the slave : Slave's address not acknowldged (begning only) - is it connected ?");
    return false;
  }
  
  else if (nRet == 3)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("Error occured wile trying to write to the slave :");
    Serial.println("Slaves's begining of address acknowledged but NACK encountered during the following writes - are all of the 10 bits of slave's address correct ? Are the writes value are meaningful to the slave ?");
    
    return false;
  }

  else if (nRet == 4)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("I2C protocol error occured wile trying to write to the slave. Is the wiring correct ? TWI Initialisation ?");
    return false;
  }
  
  //else :
  nRet = twi_readFrom(SevenBits_compat_address, destinationBuffer, readSize, true); //xxx bytes read with a Stop at the end.  
  if (nRet == 0x00)
  {
    //Incomplete reading : twi_stop() already done anyway when there is a NACK on Addr + read.
    Serial.println("Error occured wile trying to read from the slave.");
    return false;
  }

  if (nRet != readSize)
  {
    //Incomplete reading : twi_stop() already done anyway when there is a NACK on Addr + read.
    Serial.print("Unable to read all the bytes from the slave : ");
    Serial.print(nRet, DEC);
    Serial.println(" bytes read");
    
    return false;
  }

  //else :

  Serial.print("uint8_t destinationBuffer[");
  Serial.print(readSize);
  Serial.print("] = {");
  for (uint16_t i=0; i<readSize; i++)
  {
    char bufferHex[5];
    sprintf(bufferHex, "0x%02X", destinationBuffer[i]);
    Serial.write(bufferHex);
    if (i != readSize-1) Serial.print(", ");
    else Serial.print("};");
  }
  Serial.println(""); //prints a new line


  //Uncomment this to print as ASCII. 
  // It doesn't print very well, prefer using a c computer program, fopen and fwrite, and 
  // recopy previous results (looking like a C/C++ buffer declaration with the content included)
  // Then you can see it with the editor of your choice (wich is, of course, notepad++).
  /*Serial.print("destinationBuffer = ");
  for (uint16_t i=0; i<readSize; i++)
  {
    Serial.write(destinationBuffer[i]);
  }
  Serial.println(""); //prints a new line
  */
  return true;
}

void write_TI046B1_register(uint16_t TenBits_slave_address, uint8_t firstRegisterByte, uint8_t secondRegisterByte, uint8_t thirdRegisterByte, uint8_t * contentBuffer, uint16_t writeSize)
{
  uint16_t slave_address_LSB = TenBits_slave_address & 0x0FF; //0076543210 //8 LSB
  uint16_t slave_address_MSB = TenBits_slave_address & 0x300; //9800000000 //2 MSB
  
  //Put the MSB bits to the Left :
  slave_address_MSB = slave_address_MSB >> 8; //For example, 0x300 becomes 0x003
  
  //7 bits address equivalent for the begining :
  uint8_t SevenBits_compat_address = 0x78 | slave_address_MSB; //TWI library will put send those 7 bits followed by read/write bit.
  
  
  //Preparation of the 4 first bytes of the write buffer : 8LSB of the slave's address, then 3 bytes writes to call a register read.
  contentBuffer[0] = slave_address_LSB;
  contentBuffer[1] = firstRegisterByte;
  contentBuffer[2] = secondRegisterByte;
  contentBuffer[3] = thirdRegisterByte;
  
  //Execution of the 4 bytes write
  uint16_t nRet = twi_writeTo(SevenBits_compat_address, contentBuffer, 4+writeSize, 1, true); //4 bytes + content (address LSB, 3 registers bytes, and content. Wait and stop bit.
  if (nRet == 1)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("Error occured wile trying to address the slave : TWI class's buffer is too small");
    return;
  }

  else if (nRet == 2)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("Error occured wile trying to address the slave : Slave's address not acknowldged (begning only) - is it connected ?");
    return;
  }
  
  else if (nRet == 3)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("Error occured wile trying to address the slave :");
    Serial.println("Slaves's begining of address acknowledged but NACK encountered during the following writes - are all of the 10 bits of slave's address correct ? Are the writes value are meaningful to the slave ?");
    
    return;
  }

  else if (nRet == 4)
  {
    //it seems, reading TWI code, that the stop is sent anyway when there is a NACK. No need to call twi_stop().
    Serial.println("I2C protocol error occured wile trying to address the slave. Is the wiring correct ? TWI Initialisation ?");
    return;
  }
  
  //else :
  Serial.println("Write OK");
}

void Reset56(uint8_t *Buffer)
{
  //Byte 1 : 0xFF on every seen cartridge
  Buffer[0] = 0xFF;
  
  //Bytes 2 and 3 : 0xFFFF is full, 0x7FFF after 2 or 3 pages, 0x003F : 0%, 0x007F : 10~20%, 0x01FF for 40%, 0x07FF for 60%
  Buffer[1] = 0xFF;
  Buffer[2] = 0xFF;
  
  //Bytes 4 and 5 : 0xFF on every seen cartridge
  Buffer[3] = 0xFF;
  Buffer[4] = 0xFF;

  //Byte 6 : 0x1F for Full or almost full, 0x8C for empty, 0x8D for almost empty with warning sign, 0x0F seen for 30% 40% and 60%.
  Buffer [5] = 0x1F;
  
  //Byte 7 : 0x0A for starter kits, 0x00 for others : do not touch !

  //Bytes 8 to 14 : 0xFF then 0x00, 0x00... on every seen cartridge
  Buffer[7] = 0xFF;
  Buffer[8] = 0x00;
  Buffer[9] = 0x00;
  Buffer[10] = 0x00;
  Buffer[11] = 0x00;
  Buffer[12] = 0x00;
  Buffer[13] = 0x00;

  //Byte 15 and 16 : printed pages : 0x0000 for new
  Buffer[14] = 0x00;
  Buffer[15] = 0x00;

  //Bytes 17 18 19 20 : 0x00 for the first byte everywhere, then 0x??, 0x??, 0x?? depending on cartridge. Non proportionnal. 0x00000000 when new or almost new
  Buffer[16] = 0x00;
  Buffer[17] = 0x00;
  Buffer[18] = 0x00;
  Buffer[19] = 0x00;
  
  //Bytes 21 22 23 24 : 0x00 then 3 bytes depending on the cartridge model : do not touch !
  //back starter : 0x00 61 7F FF
  //cyan starter : 0x00 69 7F FF
  //magenta starter: 0x00 6D 7F FF
  //yellow starter: 0x00 63 7F FF

  //magenta 71B20M0 : 0x00 C0 00 00
  //magenta 802HME : 0x00 CD 00 00
  //cyan high yield for CS310dn : 0x00 D7 00 00
  //yellow normal yield for CS310dn : 0x00 52 00 00


  //Bytes 25 26 27 28 : 0x00 00 00 00 on every seen cartridge
  Buffer[24] = 0x00;
  Buffer[25] = 0x00;
  Buffer[26] = 0x00;
  Buffer[27] = 0x00;

  //Bytes 29 and 30 : 0x0000 when new or >= 30%, 0x01 90 (=400) or 0x011D (=285) when warning low level. Final count down rising as approaching the end ?
  Buffer[28] = 0x00;
  Buffer[29] = 0x00;

  //Bytes 31 32 33 34 35 36 37 38 39 40 : 0x 00000... on every seen cartridge
  Buffer[30] = 0x00;
  Buffer[31] = 0x00;
  Buffer[32] = 0x00;
  Buffer[33] = 0x00;
  Buffer[34] = 0x00;
  Buffer[35] = 0x00;
  Buffer[36] = 0x00;
  Buffer[37] = 0x00;
  Buffer[38] = 0x00;
  Buffer[39] = 0x00;

  //Byte 41 and 42 : 0xFFFF for starter kits, 0x0000 if not starter kit : Do not touch !

  //Byte 43 and 44 : 0x0000 on every seen cartridge
  Buffer[42] = 0x00;
  Buffer[43] = 0x00;

  //Byte 45 and 46 : 0xFFFF for starter kits, 0x0000 if not starter kit : Do not touch !

  //Byte 47 : 0x00 if not used, seen at 0x01 on 2 pages printed, seen as 0x02 if >3 pages, seen at 0x03 when almost (or completely empty)
  Buffer[46] = 0x00;

  //Byte 48 and 49 : 0x0000 on every seen cartridge
  Buffer[47] = 0x00;
  Buffer[48] = 0x00;

  //Byte 50 and 51 : 0x0000 = new, 0x0001 = few usage, 00AF = ~60%, 0x04C3 = ~40%, 0x0505 = ~30%, 030B and 04C0 = low,  0x04C4 and 0x030C : empty... so when level is low, it's imprevisible
  Buffer[49] = 0x00;
  Buffer[50] = 0x00;

  //Byte 52 53 54 : 0x00 00 00 everywhere : do not touch ?
  Buffer[51] = 0x00;
  Buffer[52] = 0x00;
  Buffer[53] = 0x00;

  uint16_t checksum = crc16(Buffer, 54);
  uint16_t checksum_MSB = (checksum & 0xFF00) >> 8;
  uint16_t checksum_LSB = (checksum & 0x00FF);

  //Byte 55 56 : CRC16
  Buffer[54] = checksum_MSB;
  Buffer[55] = checksum_LSB;
}

void Reset208(uint8_t *Buffer)
{
  //Clearing the 169 first bytes [0 to 168]
  for (uint8_t i=0; i<169; i++) Buffer[i] = 0x00;

  /*
  Uncomment this to create a new 30 bytes unique ID at [169 to 198] by mixing all of the bytes
  
  Buffer[169] = Buffer[169] + Buffer[198];
  //Buffer[169 to 198] is a 30 byte UUID that probably should be changed in the future. Let's test without, first.
  for (uint8_t i=170; i<199; i++) Buffer[i] = Buffer[i] + Buffer[i-1];
  
  //Buffer[199 to 205] is 7 bytes of 0x00
  for (uint8_t i=199; i<206; i++) Buffer[i] = 0x00;
  */

  uint16_t checksum = crc16(Buffer, 206);
  uint16_t checksum_MSB = (checksum & 0xFF00) >> 8;
  uint16_t checksum_LSB = (checksum & 0x00FF);

  //Byte 206 207 : CRC16
  Buffer[206] = checksum_MSB;
  Buffer[207] = checksum_LSB;
}

uint8_t BufferPlus4[1024 + 4]; //Add 4 bytes on left for the one shot write procedure (LSB and 3 bytes of registers address before data).
uint8_t *Buffer = &(BufferPlus4[4]); //Same buffer begining after the 4 first bytes, for read, modification, and write back after.

void ResetBlack()
{
  bool bRet = read_TI046B1_register(0x001, 0x01, 0x40, 0x04, Buffer, 12);
  if (bRet == false) return;
  Buffer[12] = 0x00;
  Serial.print("Black serial number : ");
  Serial.println((char*)Buffer);
  
  //Black, 56 bytes, first one
  Serial.println("Black, (0x001), 56 bytes register, first one");
  bRet = read_TI046B1_register(0x001, 0x01, 0x20, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x001, 0x02, 0x20, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x001, 0x01, 0x20, 0x00, Buffer, 56); //After
  Serial.println("");
  
  //Black, 56 bytes, second one
  Serial.println("Black, (0x001), 56 bytes register, second one");
  bRet = read_TI046B1_register(0x001, 0x01, 0x58, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x001, 0x02, 0x58, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x001, 0x01, 0x58, 0x00, Buffer, 56); //After
  Serial.println("");

  //Clear the 208 bytes register, for Black
  Serial.println("Black, (0x001), 208 bytes register");
  bRet = read_TI046B1_register(0x001, 0x01, 0x90, 0x00, Buffer, 208);
  if (bRet == false) return;
  Reset208(Buffer);
  write_TI046B1_register(0x001, 0x02, 0x90, 0x00, BufferPlus4, 208);
  read_TI046B1_register(0x001, 0x01, 0x90, 0x00, Buffer, 208);
  Serial.println("");
}

void ResetBlue()
{
  bool bRet = read_TI046B1_register(0x002, 0x01, 0x40, 0x04, Buffer, 12);
  if (bRet == false) return;
  Buffer[12] = 0x00;
  Serial.print("Blue serial number : ");
  Serial.println((char*)Buffer);

  //Cyan, 56 bytes, first one
  Serial.println("Blue, (0x002), 56 bytes register, first one");
  bRet = read_TI046B1_register(0x002, 0x01, 0x20, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x002, 0x02, 0x20, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x002, 0x01, 0x20, 0x00, Buffer, 56); //After
  Serial.println("");

  //Cyan, 56 bytes, second one
  Serial.println("Blue, (0x002), 56 bytes register, second one");
  bRet = read_TI046B1_register(0x002, 0x01, 0x58, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x002, 0x02, 0x58, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x002, 0x01, 0x58, 0x00, Buffer, 56); //After
  Serial.println("");

  //Clear the 208 bytes register, for Cyan
  Serial.println("Blue, (0x002), 208 bytes register");
  bRet = read_TI046B1_register(0x002, 0x01, 0x90, 0x00, Buffer, 208);
  if (bRet == false) return;
  Reset208(Buffer);
  write_TI046B1_register(0x002, 0x02, 0x90, 0x00, BufferPlus4, 208);
  read_TI046B1_register(0x002, 0x01, 0x90, 0x00, Buffer, 208);
  Serial.println("");
}

void ResetMagenta()
{
  bool bRet = read_TI046B1_register(0x003, 0x01, 0x40, 0x04, Buffer, 12);
  if (bRet == false) return;
  Buffer[12] = 0x00;
  Serial.print("Magenta serial number before : ");
  Serial.println((char*)Buffer);
  
  //Magenta, 56 bytes, first one
  Serial.println("Magenta, (0x003), 56 bytes register, first one");
  bRet = read_TI046B1_register(0x003, 0x01, 0x20, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x003, 0x02, 0x20, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x003, 0x01, 0x20, 0x00, Buffer, 56); //After
  Serial.println("");

  //Magenta, 56 bytes, second one
  Serial.println("Magenta, (0x003), 56 bytes register, second one");
  bRet = read_TI046B1_register(0x003, 0x01, 0x58, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x003, 0x02, 0x58, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x003, 0x01, 0x58, 0x00, Buffer, 56); //After
  Serial.println("");
  
  //Clear the 208 bytes register, for Magenta
  Serial.println("Magenta, (0x003), 208 bytes register");
  bRet = read_TI046B1_register(0x003, 0x01, 0x90, 0x00, Buffer, 208);
  if (bRet == false) return;
  Reset208(Buffer);
  write_TI046B1_register(0x003, 0x02, 0x90, 0x00, BufferPlus4, 208);
  read_TI046B1_register(0x003, 0x01, 0x90, 0x00, Buffer, 208);
  Serial.println("");
}

void ResetYellow()
{
  bool bRet = read_TI046B1_register(0x004, 0x01, 0x40, 0x04, Buffer, 12);
  if (bRet == false) return;
  Buffer[12] = 0x00;
  Serial.print("Yellow serial number : ");
  Serial.println((char*)Buffer);
    
  //Yellow, 56 bytes, first one
  Serial.println("Yellow, (0x004), 56 bytes register, first one");
  bRet = read_TI046B1_register(0x004, 0x01, 0x20, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x004, 0x02, 0x20, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x004, 0x01, 0x20, 0x00, Buffer, 56); //After
  Serial.println("");
  
  //Yellow, 56 bytes, second one
  Serial.println("Yellow, (0x004), 56 bytes register, second one");
  bRet = read_TI046B1_register(0x004, 0x01, 0x58, 0x00, Buffer, 56); //Before
  if (bRet == false) return;
  Reset56(Buffer);
  write_TI046B1_register(0x004, 0x02, 0x58, 0x00, BufferPlus4, 56);
  read_TI046B1_register(0x004, 0x01, 0x58, 0x00, Buffer, 56); //After
  Serial.println("");
  
  //Clear the 208 bytes register, for Yellow
  Serial.println("Yellow, (0x004), 208 bytes register");
  bRet = read_TI046B1_register(0x004, 0x01, 0x90, 0x00, Buffer, 208);
  if (bRet == false) return;
  Reset208(Buffer);
  write_TI046B1_register(0x004, 0x02, 0x90, 0x00, BufferPlus4, 208);
  read_TI046B1_register(0x004, 0x01, 0x90, 0x00, Buffer, 208);
  Serial.println("");
}


void playOnce()
{
  ResetBlack();
  Serial.println("-------------------------------------------");
  ResetBlue();
  Serial.println("-------------------------------------------");
  ResetMagenta();
  Serial.println("-------------------------------------------");
  ResetYellow();
}

bool played = false;

void loop()
{
  if (played == false)
  {
    playOnce();
    played = true;
  }
}

